// -*- C++ -*-
/*!
 * @file  CorbaPort.h
 * @brief CorbaPort class
 * @date  $Date: 2007/01/04 00:45:21 $
 * @author Noriaki Ando <n-ando@aist.go.jp>
 *
 * Copyright (C) 2006
 *     Noriaki Ando
 *     Task-intelligence Research Group,
 *     Intelligent Systems Research Institute,
 *     National Institute of
 *         Advanced Industrial Science and Technology (AIST), Japan
 *     All rights reserved.
 *
 * $Id: CorbaPort.h,v 1.1 2007/01/04 00:45:21 n-ando Exp $
 *
 */

/*
 * $Log: CorbaPort.h,v $
 * Revision 1.1  2007/01/04 00:45:21  n-ando
 * CORBA serivce provider/consumer Port implementation.
 *
 *
 */


#ifndef CorbaPort_h
#define CorbaPort_h

#include <rtm/PortBase.h>
#include <rtm/CorbaConsumer.h>
#include <rtm/NVUtil.h>

namespace RTC
{
  /*!
   * @if jp
   * @class CorbaPort
   * @brief RT R|[lg CORBA service/consumer p Port
   *
   * CorbaPort  RT R|[lgɂāA[U` CORBA IuWFNg
   * T[rXуRV[}񋟂 Port łB
   * <p>
   * RT R|[lǵAPort ă[U` CORBA T[rX
   * 邱ƂłA RT Service (Provider) ƌĂԁB
   * ܂A RT R|[lg̃T[rX𗘗p邽߂ CORBA IuWFNg
   * ̃v[Xz_񋟂邱ƂłA RT Service Consumer ƌĂԁB
   * <p>
   * CorbaPort ͔Cӂ̐ Provider  Consumer Ǘ邱ƂłA
   * Port mڑۂɑΉ Provider  Consumer K؂Ɋ֘At
   * ƂłB
   * <p>
   * CorbaPort ͒ʏȉ̂悤ɗpB
   *
   * <pre>
   * RTC::CorbaPort m_port0; // Port ̐錾
   *
   * MyService_impl m_mysvc0; //  Port 񋟂 Serivce Provider
   * RTC::CorbaConsumer<YourService> m_cons0; //  Port  Consumer
   *
   * // Service Provider  Port ɓo^
   * m_port0.registerProvider("MyService0", "Generic", m_mysvc0);
   * // Service Consumer  Port ɓo^
   * m_port0.registerConsumer("YourService0", "Generic", m_cons0 );
   *
   * // connect sꂽ
   *
   * m_cons0->your_service_function(); // YourService ̊֐R[
   *
   * // connect ꂽ ʂ̃R|[lgɂ
   * m_cons1->my_service_function(); // MyService ̊֐R[
   * </pre>
   *
   * ̂悤ɁA񋟂 Service Provider  registerProvider() œo^
   * 邱ƂɂÃR|[lg痘p\ɂAA
   * p Service Consumer  registerConsumer() œo^邱Ƃɂ
   * ̃R|[lg Service R|[lgŗp\ɂ邱Ƃ
   * łB
   *
   * @else
   * @class CorbaPort
   * @brief RT Conponent CORBA service/consumer Port
   *
   * CorbaPort is an implementation of the Port of RT-Component's that provides
   * user-defined CORBA Object Service and Consumer.
   * <p>
   * RT-Component can provide user-defined CORBA serivces, which is called
   * RT-Serivce (Provider), through the Ports.
   * RT-Component can also provide place-holder, which is called RT-Serivce
   * Consumer, to use other RT-Component's service.
   * <p>
   * The CorbaPort can manage any number of Providers and Consumers, can
   * associate Consumers with correspondent Providers when establishing
   * connection among Ports.
   * <p>
   * Usually, CorbaPort is used like the following.
   *
   * <pre>
   * RTC::CorbaPort m_port0; // declaration of Port
   *
   * MyService_impl m_mysvc0; // Serivce Provider that is provided by the Port
   * RTC::CorbaConsumer<YourService> m_cons0; // Consumer of the Port
   *
   * // register Service Provider to the Port
   * m_port0.registerProvider("MyService0", "Generic", m_mysvc0);
   * // register Service Consumer to the Port
   * m_port0.registerConsumer("YourService0", "Generic", m_cons0 );
   *
   * // after connect established
   *
   * m_cons0->your_service_function(); // call a YourService's function
   *
   * // in another component that is connected with the Port
   * m_cons1->my_service_function(); // call a MyService's function
   * </pre>
   *
   * Registering Service Provider by registerProvider(), it can be used from
   * other RT-Components.
   * Registering Service Consumer by registerConsumer(), other RT-Component's
   * services can be used through the consumer object.
   *
   * @endif
   */
  class CorbaPort
    : public PortBase
  {
  public:
    /*!
     * @if jp
     * @brief RXgN^
     *
     * @param name Port ̖O
     *
     * @else
     *
     * @brief Constructor
     *
     * @param name The name of Port 
     *
     * @endif
     */
    CorbaPort(const char* name);


    /*!
     * @if jp
     *
     * @brief fXgN^
     *
     * @else
     *
     * @brief Destructor
     *
     * @endif
     */
    virtual ~CorbaPort();


    /*!
     * @if jp
     *
     * @brief Provider o^
     *
     *  Port ɂĒ񋟂T[og Port ɑ΂ēo^B
     * T[oǵAŗ^ instance_name, type_name A
     * T[ogg̃CX^Xу^CvƂāAT[og
     * ֘AtB
     *
     * @param instance_name T[og̃CX^X
     * @param type_name T[og̃^Cv
     * @param provider CORBA T[og
     * @return ɓ instance_name o^Ă false ԂB
     *
     * @else
     *
     * @brief Register provider
     *
     * This operation registers a servant, which is provided in this Port,
     * to the Port. The servant is associated with "instance_name" and
     * "type_name" as the instance name of the servant and as the type name
     * of the servant.
     *
     * @param instance_name T[og̃CX^X
     * @param type_name T[og̃^Cv
     * @param provider CORBA T[og
     * @return ɓ instance_name o^Ă false ԂB
     *
     * @endif
     */
    bool registerProvider(const char* instance_name, const char* type_name,
			  PortableServer::RefCountServantBase& provider);


    /*!
     * @if jp
     *
     * @brief Consumer o^
     *
     *  Port vT[rX̃v[Xz_łRV[}
     * (Consumer) o^B
     * Consumer ֘AtT[rX̃CX^Xу^CvƂāA
     *  instance_name, type_name  Consumer g^邱ƂɂA
     * ł炪֘AtB
     * Port Ԃ̐ڑ (connect)  ɂ́A instance_name, type_name 
     * T[rX Port  (Provide) ĂꍇÃT[rX
     * IuWFNgQƂI Consumer ɃZbgB
     *
     * @param instance_name Consumer vT[rX̃CX^X
     * @param type_name Consumer vT[rX̃^Cv
     * @param consumer CORBA T[rXRV[}
     * @return ɓ instance_name o^Ă false ԂB
     *
     * @else
     *
     * @brief Register consumer
     *
     * This operation registers a consumer, which requiers a service,
     * to the other Port. The consumer is associated with "instance_name" and
     * "type_name" as the instance name of the service and as the type name
     * of the service that is required.
     *
     * @param instance_name An instance name of the service required
     * @param type_name An type name of the service required
     * @param consumer CORBA service consumer
     * @return False would be returned if the same instance_name is registered
     *
     * @endif
     */
    bool registerConsumer(const char* instance_name, const char* type_name,
			  CorbaConsumerBase& consumer);


  protected:
    /*!
     * @if jp
     *
     * @brief Interface J
     *
     *  PortL Provider Ɋւ ConnectorProfile::properties
     * ɑB
     * ́ANVList name  value ƂĈȉ̂̂i[B
     *
     * - port.<type_name>.<instance_name>: <CORBA::Object_ptr>
     *
     * ŁA
     * - <type_name>: PortInterfaceProfile::type_name
     * - <instance_name>: PortInterfaceProfile::instance_name
     * łB
     * ConnectorProfile::properties ł́A .(hbg)\LŁA
     * NameValue ̃L[ƂĂBāA
     *
     * <pre>
     *  PortInterfaceProfile
     *  {
     *    instance_name = "PA10_0";
     *    type_name     = "Manipulator";
     *    polarity      = PROVIDED;
     *  }
     *</pre>
     *
     * Ȃ΁A
     *
     * <pre>
     * NameValue = { "port.Manipulator.PA10_0": <Object reference> }
     * </pre>
     *
     * Ƃl ConnectorProfile::properties Ɋi[Ã|[gɑ΂
     * `BB Port ł̃C^[tF[Xgp Consumer 
     * ݂΁AConnectorProfile 炱̃L[IuWFNgt@X
     * 擾炩̌`ŎgpB
     *
     * @else
     *
     * @brief Publish interface information
     *
     * @endif
     */
     virtual ReturnCode_t
     publishInterfaces(ConnectorProfile& connector_profile);


    /*!
     * @if jp
     *
     * @brief Interface ɐڑ
     *
     *  PortL Consumer ɓK Provider Ɋւ
     * ConnectorProfile::properties 璊o Consumer ɃIuWFNgQ
     * ZbgB
     *
     * AConsumer 
     * <pre>
     *  PortInterfaceProfile
     *  {
     *    instance_name = "PA10_0";
     *    type_name     = "Manipulator";
     *    polarity      = REQUIRED;
     *  }
     *</pre>
     * Ƃēo^Ă΁A Port 
     * <pre>
     *  PortInterfaceProfile
     *  {
     *    instance_name = "PA10_0";
     *    type_name     = "Manipulator";
     *    polarity      = PROVIDED;
     *  }
     * </pre> 
     * Ƃēo^Ă Serivce Provider ̃IuWFNgQƂTA
     * Consumer ɃZbgB
     * ۂɂ́AConnectorProfile::properties 
     * <pre>
     * NameValue = { "port.Manipulator.PA10_0": <Object reference> }
     * </pre>
     * Ƃēo^Ă NameValue TÃIuWFNgQƂ
     * Consumer ɃZbgB
     *
     * @else
     *
     * @brief Subscribe interfaces
     *
     * @endif
     */
     virtual ReturnCode_t
     subscribeInterfaces(const ConnectorProfile& connector_profile);


    /*!
     * @if jp
     *
     * @brief Interface ւ̐ڑ
     *
     * ^ꂽ ConnectorProfile Ɋ֘A Consumer ɃZbgꂽ
     * ׂĂ Object ڑB
     *
     * @else
     *
     * @brief Unsubscribe interfaces
     *
     * @endif
     */
     virtual void
     unsubscribeInterfaces(const ConnectorProfile& connector_profile);


  private:
    /*!
     * @if jp
     * @brief Provider ̏i[ NVList
     * @else
     * @brief NVList to be stored Providers' information
     * @endif
     */
    NVList m_providers;


    /*!
     * @if jp
     * @brief Consumer ̏i[\
     * @else
     * @brief Consumer inforamtion struct
     * @endif
     */
    struct Consumer
    {
      Consumer(const char* _instance_name, const char* _type_name,
	       CorbaConsumerBase& _cons)
	: name(std::string("port.")
	       + _type_name
	       + std::string(".")
	       + _instance_name),
	  consumer(_cons)
      {}
      Consumer(const Consumer& cons)
	: name(cons.name), consumer(cons.consumer)
      {	
      }
      Consumer operator=(const Consumer& _cons)
      {
	Consumer cons(_cons);
	return cons;
      }
      std::string name;
      CorbaConsumerBase& consumer;
    };
    std::vector<Consumer> m_consumers;

    // functors
    /*!
     * @if jp
     * @brief ConnectorProfile  Consuemr ̔rIuWFNgQƂ
     *        Zbg邽߂ Functor
     * @else
     * @brief Subscription mutching functor for Consumer
     * @endif
     */
    struct subscribe
    {
      subscribe(std::vector<Consumer>& cons)
	: m_cons(cons), m_len(cons.size())
	{
	}
	
	void operator()(const SDOPackage::NameValue& nv)
	{
	  for (CORBA::ULong i = 0; i < m_len; ++i)
	    {
	      std::string name(nv.name);
	      if (m_cons[i].name == name)
		{
		  CORBA::Object_ptr obj;
		  if (nv.value >>= CORBA::Any::to_object(obj))
		    {
		      m_cons[i].consumer.setObject(obj);
		    }
		}
	}
	}
	std::vector<Consumer> m_cons;
	CORBA::ULong m_len;
    };


    /*!
     * @if jp
     * @brief Consumer ̃IuWFNg邽߂ Functor
     * @else
     * @brief Unsubscription functor for Consumer
     * @endif
     */
    struct unsubscribe
    {
      unsubscribe(std::vector<Consumer>& cons)
	: m_cons(cons), m_len(cons.size())
      {
      }
      
      void operator()(const SDOPackage::NameValue& nv)
      {
	for (CORBA::ULong i = 0; i < m_len; ++i)
	  {
	    std::string name(nv.name);
	    if (m_cons[i].name == name)
	      {
		m_cons[i].consumer.releaseObject();
	      }
	  }
      }
      std::vector<Consumer> m_cons;
      CORBA::ULong m_len;
    };
    
  };
};
#endif // CorbaPort_h
