// -*- C++ -*-
/*!
 * @file Factory.h
 * @brief RTComponent factory class
 * @date $Date: 2006/11/06 01:28:36 $
 * @author Noriaki Ando <n-ando@aist.go.jp>
 *
 * Copyright (C) 2003-2005
 *     Task-intelligence Research Group,
 *     Intelligent Systems Research Institute,
 *     National Institute of
 *         Advanced Industrial Science and Technology (AIST), Japan
 *     All rights reserved.
 *
 * $Id: Factory.h,v 1.5 2006/11/06 01:28:36 n-ando Exp $
 *
 */

/*
 * $Log: Factory.h,v $
 * Revision 1.5  2006/11/06 01:28:36  n-ando
 * Now the "instance_name" is set to the component at creation time.
 *
 * Revision 1.4  2006/10/25 17:36:00  n-ando
 * Classes were renamed, and class reference manual was described.
 *
 * Revision 1.3  2006/10/24 14:23:45  n-ando
 * Renamed RtcFactory.h to Factory.h
 *
 * Revision 1.2  2005/05/16 06:06:30  n-ando
 * - RtcFactoryBase, RtcFactoryCXX classes were DLL exported for Windows port.
 *
 * Revision 1.1.1.1  2005/05/12 09:06:18  n-ando
 * Public release.
 *
 *
 */

#ifndef RtcFactory_h
#define RtcFactory_h
#define EXPORTS
#include "rtm/Properties.h"
#include "rtm/RTObject.h"
#include <rtm/NumberingPolicy.h>


namespace RTC 
{
  typedef RTObject_impl RtcBase;
  class Manager;

  typedef RtcBase* (*RtcNewFunc)(Manager* manager);
  typedef void (*RtcDeleteFunc)(RtcBase* rtc);

  template <class _New>
  RtcBase* Create(Manager* manager)
  {
    return new _New(manager);
  }

  template <class _Delete>
  void Delete(RtcBase* rtc)
  {
    delete rtc;
  }


  /*!
   * @if jp
   *
   * @class RtcFactoryBase
   *
   * @brief RtcFactoryBase NX
   * 
   * R|[lgt@Ng̊NXB
   *
   * @else
   *
   * @class RtcFactoryBase
   *
   * @brief RtcFactoryBase base class
   *
   * RTComponent factory base class.
   *
   * @endif
   */
  class EXPORTS FactoryBase
  {
  public:
    /*!
     * @if jp
     *
     * @brief RtcFactoryBase NXRXgN^
     *
     * RtcFactoryBase NX̃RXgN^B
     *
     * @param profile R|[lg̃vt@C
     *
     * @else
     *
     * @brief RtcFactoryBase class constructor.
     *
     * RtcFactoryBase class constructor.
     *
     * @param profile component profile
     *
     * @endif
     */
    FactoryBase(const Properties& profile);
    virtual ~FactoryBase();
    
    /*!
     * @if jp
     *
     * @brief R|[lg̐
     *
     * Python Ŏꂽ RTComponent ̃CX^X𐶐B
     * z֐B
     *
     * @param mgr RtcManager ւ̃|C^
     *
     * @else
     *
     * @brief Create component
     *
     * Create component implemented in Python.
     * Pure virtual method.
     *
     * @param mgr pointer to RtcManager
     *
     * @endif
     */
    virtual RtcBase* create(Manager* mgr) = 0;
    
    /*!
     * @if jp
     *
     * @brief R|[lg̔j
     *
     * RTComponent ̃CX^XjB
     * z֐B
     *
     * @param comp RtcBase ւ̃|C^
     *
     * @else
     *
     * @brief Destroy component
     *
     * Destroy component instance
     * Pure virtual method.
     *
     * @param comp pointer to RtcBase
     *
     * @endif
     */
    virtual void destroy(RtcBase* comp) = 0;
    
    /*!
     * @if jp
     *
     * @brief R|[lgvt@C̎擾
     *
     * R|[lg̃vt@C擾
     *
     * @else
     *
     * @brief Get component profile
     *
     * Get component profile.
     *
     * @endif
     */
    virtual Properties& profile();
    
    /*!
     * @if jp
     *
     * @brief ݂̃CX^X
     *
     * R|[lǧ݂̃CX^X擾B
     *
     * @else
     *
     * @brief Get number of component instances
     *
     * Get number of current component instances.
     *
     * @endif
     */
    virtual int number();
    
  protected:
    /*!
     * @if jp
     * @brief R|[lg̃vt@C
     * @else
     * @brief Component profile
     * @endif
     */
    Properties m_Profile;
    /*!
     * @if jp
     * @brief ݂̃CX^X
     * @else
     * @brief Number of current component instances.
     * @endif
     */
    int m_Number;
  };
  
  
  
  /*!
   * @if jp
   *
   * @class RtcFactoryCXX
   *
   * @brief RtcFactoryCXX NX
   * 
   * C++pR|[lgt@NgNXB
   *
   * @else
   *
   * @class RtcFactoryCXX
   *
   * @brief RtcFactoryCXX class
   *
   * RTComponent factory class for C++.
   *
   * @endif
   */
  class EXPORTS FactoryCXX
    : public FactoryBase
  {
  public:
    /*!
     * @if jp
     *
     * @brief RtcFactoryCXX NXRXgN^
     *
     * RtcFactoryCXX NX̃RXgN^B
     * vt@CA֐ւ̃|C^Aj֐ւ̃|C^ɎA
     * C++ ŎꂽR|[lg̃t@NgNX𐶐B
     *
     * @param profile R|[lg̃vt@C
     * @param new_func R|[lg̐֐ւ̃|C^
     * @param delete_func R|[lg̔j֐ւ̃|C^
     *
     * @else
     *
     * @brief RtcFactoryCXX class constructor.
     *
     * RtcFactoryCXX class constructor.
     * Create component factory class with three arguments:
     * component profile, function pointer to object create function and
     * object delete function.
     *
     * @param profile Component profile
     * @param new_func Pointer to component create function
     * @param delete_func Pointer to component delete function
     *
     * @endif
     */
    FactoryCXX(const Properties& profile,
	       RtcNewFunc new_func,
	       RtcDeleteFunc delete_func,
	       NumberingPolicy* policy = new DefaultNumberingPolicy());
    
    /*!
     * @if jp
     *
     * @brief R|[lg̐
     *
     * Python Ŏꂽ RTComponent ̃CX^X𐶐B
     *
     * @param mgr RtcManager ւ̃|C^
     *
     * @else
     *
     * @brief Create component
     *
     * Create component implemented in Python.
     *
     * @param mgr pointer to RtcManager
     *
     * @endif
     */
    virtual RtcBase* create(Manager* mgr);
    
    /*!
     * @if jp
     *
     * @brief R|[lg̔j
     *
     * RTComponent ̃CX^XjB
     *
     * @param comp RtcBase ւ̃|C^
     *
     * @else
     *
     * @brief Destroy component
     *
     * Destroy component instance
     *
     * @param comp pointer to RtcBase
     *
     * @endif
     */
    virtual void destroy(RtcBase* comp);
    
  protected:
    /*!
     * @if jp
     * @brief R|[lgIuWFNg֐ւ̃|C^
     * @else
     * @brief The pointer to component object create function
     * @endif
     */
    RtcNewFunc m_New;
    
    /*!
     * @if jp
     * @brief R|[lgIuWFNgj֐ւ̃|C^
     * @else
     * @brief The pointer to component object delete function
     * @endif
     */
    RtcDeleteFunc m_Delete;

    NumberingPolicy* m_policy;
  };
};


#endif // RtcFactory_h
